@@ -2,7 +2,7 @@
 #
 # The fisx library for X-Ray Fluorescence
 #
-# Copyright (c) 2014-2016 European Synchrotron Radiation Facility
+# Copyright (c) 2014-2018 European Synchrotron Radiation Facility
 #
 # This file is part of the fisx X-ray developed by V.A. Sole
 #
@@ -32,6 +32,7 @@
 #include <cmath>
 #include <sstream>
 #include <algorithm>
+#include <cstdlib>
 #include "fisx_elements.h"
 
 namespace fisx
@@ -40,7 +41,17 @@ namespace fisx
 
 const std::string Elements::defaultDataDir()
 {
-    return FISX_DATA_DIR;
+    char *path;
+
+    path = std::getenv("FISX_DATA_DIR");
+    if (path != NULL)
+    {
+        return std::string(path);
+    }
+    else
+    {
+        return FISX_DATA_DIR;
+    }
 }
 
 Elements::Elements(std::string epdl97Directory, std::string bindingEnergiesFileName, std::string crossSectionsFile)
@@ -59,7 +70,14 @@ Elements::Elements(std::string epdl97Directory, std::string bindingEnergiesFileN
 Elements::Elements(std::string epdl97Directory)
 {
     // pure EPDL97 initialization
-    this->initialize(epdl97Directory, "");
+    if (epdl97Directory.size() < 1)
+    {
+        this->initialize(Elements::defaultDataDir(), "");
+    }
+    else
+    {
+        this->initialize(epdl97Directory, "");
+    }
 }
 
 Elements::Elements(std::string directoryName, short pymca)
@@ -701,6 +719,7 @@ void Elements::setMassAttenuationCoefficients(const std::string & name,
     std::map<std::string, std::vector<double> > massAttenuationCoefficients;
     std::map<std::string, std::vector<double> >::iterator it;
     std::map<std::string, std::pair<double, int> > extractedEdgeEnergies;
+    std::vector<std::vector<double>::size_type> shellIndexList;
     std::string shellList[10] = {"K", "L1", "L2", "L3", "M1", "M2", "M3", "M4", "M5", "all other"};
     Element *element;
     std::vector<double> newEnergyGrid;
@@ -757,14 +776,63 @@ void Elements::setMassAttenuationCoefficients(const std::string & name,
         j--;
     }
 
+    // partial cross sections
+    // we extract the edges from supplied mass attenuation coefficients
+    // the result contains the indices of the first instance of the energy
+    extractedEdgeEnergies = (*element).extractEdgeEnergiesFromMassAttenuationCoefficients(energy, \
+                                                                                          photoelectric);
+
+    shellIndexList.clear();
+    for (shellIndex = 0; shellIndex < 10; shellIndex++)
+    {
+        shell = shellList[shellIndex];
+        if (extractedEdgeEnergies.find(shell) != extractedEdgeEnergies.end())
+        {
+            // we have the index prior to starting to excite that shell but that index corresponds to
+            // the old, non extended, energy grid
+            idx = extractedEdgeEnergies[shell].second;
+            shellIndexList.push_back(idx);
+        }
+    }
+
     newEnergyGrid.resize(j + energy.size());
     for (i = 0; i < j; i++)
     {
         newEnergyGrid[i] = massAttenuationCoefficients["energy"][i];
     }
-    for (i=j; i < newEnergyGrid.size() ; i++)
+
+    for (i = j; i < newEnergyGrid.size() ; i++)
     {
         newEnergyGrid[i] = energy[i - j];
+        // Check compatibility
+        for (shellIndex = 0; shellIndex < 10; shellIndex++)
+        {
+            shell = shellList[shellIndex];
+            if (extractedEdgeEnergies.find(shell) != extractedEdgeEnergies.end())
+            {
+                if ((i - j - 1) == static_cast<unsigned int>(extractedEdgeEnergies[shell].second))
+                {
+                    // We are just above an edge according to the input mass attenuation coefficients
+                    // We have to check if that energy is compatible with the binding energy.
+                    // If the energy was below the binding energy of that shell, we have to make sure the point is considered
+                    // (Problem with excitation of Pb at 14 keV when using XCOM cross sections)
+                    std::map<std::string, double> tmpMap = this->getBindingEnergies(name);
+                    if (extractedEdgeEnergies[shell].first < tmpMap[shell])
+                    {
+                        // I should interpolate in the supplied grid to get the proper value
+                        newEnergyGrid[i] = tmpMap[shell];
+                        continue;
+                    }
+                }
+            }
+        }
+        if (j > 0)
+        {
+            if (newEnergyGrid[i] < newEnergyGrid[i - 1])
+            {
+                newEnergyGrid[i] = newEnergyGrid[i - 1] + 0.000001;
+            }
+        }
     }
 
     // we have got the extended energy grid
@@ -784,15 +852,20 @@ void Elements::setMassAttenuationCoefficients(const std::string & name,
         throw std::runtime_error(msg);
     }
 
+/*
     // test the energies
     for (i = j; i < newEnergyGrid.size(); i++)
     {
-        if (fabs(massAttenuationCoefficients["energy"][i] - energy[i - j]) > 1.0e-10)
+        if (fabs(massAttenuationCoefficients["energy"][i] - energy[i - j]) > 0.010)
         {
+            std::cout << "Inconsistent energy grid for element " << name << std::endl;
+            std::cout << " Energy = " << massAttenuationCoefficients["energy"][i] << std::endl;
+            std::cout << " Input e = " << energy[i - j] << std::endl;
             msg = "Inconsistent energy grid for element " + name;
-            throw std::runtime_error(msg);
+            //throw std::runtime_error(msg);
         }
     }
+*/
 
     // coherent
     for (i = j; i < newEnergyGrid.size(); i++)
@@ -1477,6 +1550,7 @@ std::map<std::string, double> Elements::parseFormula(const std::string & formula
     }
 
     // look for parenthesis
+    parsingKey = false;
     spacesPresent = false;
     for(i=0; i < formula.size(); i++)
     {
@@ -1789,6 +1863,11 @@ std::string Elements::toString(const double& number)
     return s;
 }
 
+std::map<std::string, double> Elements::getEmittedXRayLines(const std::string & elementName, const double & energy) const
+{
+    return this->getElement(elementName).getEmittedXRayLines(energy);
+}
+
 const std::map<std::string, double> & Elements::getNonradiativeTransitions(const std::string & elementName, \
                                                                      const std::string & subshell) const
 {
@@ -1830,6 +1909,18 @@ std::map<std::string,std::map<std::string, double> > Elements::getEscape( \
     std::string tmpString;
     double intrinsicEfficiency;
 
+    if (this->isEscapeCacheCompatible(composition,energyThreshold,intensityThreshold, \
+                                        nThreshold , alphaIn, thickness))
+    {
+        std::map< double, std::map<std::string,std::map<std::string, double> > >::const_iterator it;
+        it = this->detectorEscapeCache.find(energy);
+        if ( it != this->detectorEscapeCache.end())
+        {
+            // std::cout << "USING CACHE" <<  energy << std::endl;
+            return it->second;
+        }
+    }
+
     if (alphaIn == 90.)
         sinAlphaIn = 1.0;
     else
@@ -1844,7 +1935,16 @@ std::map<std::string,std::map<std::string, double> > Elements::getEscape( \
     result.clear();
 
     if (thickness > 0.0)
+    {
         intrinsicEfficiency = (1.0 - std::exp(-muIncident * thickness / sinAlphaIn));
+    }
+    else
+    {
+        // this is just to get rid of a not very clever non-initialized warning because
+        // if thickness is not bigger than 0.0 intrinsicEfficiency will not be used in
+        // any case.
+        intrinsicEfficiency = 1.0;
+    }
     for (c_it = composition.begin(); c_it != composition.end(); c_it++)
     {
         element = c_it->first;
@@ -1853,15 +1953,15 @@ std::map<std::string,std::map<std::string, double> > Elements::getEscape( \
         std::map<std::string, std::map<std::string, double> >::const_iterator it;
         std::map<std::string, double>::const_iterator mapIt;
         // factor uses mu["shell photoelectric"] / mu["total photoelectric"]
-        double factor;
+        // double factor;
         // rate uses mu["shell photoelectric"]
         // It is the product factor * mu["total photoelectric"]
         double rate;
         double fluorescentEnergy;
         for ( it = excitationFactors.begin(); it != excitationFactors.end(); ++it)
         {
-            mapIt = it->second.find("factor");
-            factor = mapIt->second;
+            // mapIt = it->second.find("factor");
+            // factor = mapIt->second;
             mapIt = it->second.find("rate");
             rate = mapIt->second;
             mapIt = it->second.find("energy");
@@ -1878,7 +1978,7 @@ std::map<std::string,std::map<std::string, double> > Elements::getEscape( \
                     // and not per incident photon.
                     // It is not a correct approximation, but this avoids the calculation
                     // of exponential integral functions.
-                    rate /= (1.0 - std::exp(-muIncident * thickness / sinAlphaIn));
+                    rate /= intrinsicEfficiency;
                 }
                 result[tmpString] ["rate"] = rate;
                 result[tmpString] ["energy"] = energy - fluorescentEnergy;
@@ -1970,4 +2070,191 @@ void Elements::emptyElementCascadeCache(const std::string & elementName)
         throw std::invalid_argument("Invalid element: " + elementName);
 }
 
+void Elements::fillCache(const std::string & elementName, const std::vector< double> & energy)
+{
+    std::map<std::string, int>::const_iterator it;
+    int i;
+    if (this->isElementNameDefined(elementName))
+    {
+        it = this->elementDict.find(elementName);
+        i = it->second;
+        return this->elementList[i].fillCache(energy);
+    }
+    else
+        throw std::invalid_argument("Invalid element: " + elementName);
+}
+
+void Elements::updateCache(const std::string & elementName, const std::vector< double> & energy)
+{
+    std::map<std::string, int>::const_iterator it;
+    int i;
+    if (this->isElementNameDefined(elementName))
+    {
+        it = this->elementDict.find(elementName);
+        i = it->second;
+        return this->elementList[i].updateCache(energy);
+    }
+    else
+        throw std::invalid_argument("Invalid element: " + elementName);
+}
+
+
+void Elements::setCacheEnabled(const std::string & elementName, const int & flag)
+{
+    std::map<std::string, int>::const_iterator it;
+    int i;
+    if (this->isElementNameDefined(elementName))
+    {
+        it = this->elementDict.find(elementName);
+        i = it->second;
+        return this->elementList[i].setCacheEnabled(flag);
+    }
+    else
+        throw std::invalid_argument("Invalid element: " + elementName);
+}
+
+void Elements::clearCache(const std::string & elementName)
+{
+    std::map<std::string, int>::const_iterator it;
+    int i;
+    if (this->isElementNameDefined(elementName))
+    {
+        it = this->elementDict.find(elementName);
+        i = it->second;
+        return this->elementList[i].clearCache();
+    }
+    else
+        throw std::invalid_argument("Invalid element: " + elementName);
+}
+
+void Elements::clearEscapeCache(void)
+{
+    this->detectorEscapeCache.clear();
+}
+
+
+bool Elements::isEscapeCacheCompatible(\
+                                        const std::map<std::string, double> & composition,
+                                        const double & energyThreshold, \
+                                        const double & intensityThreshold, \
+                                        const int & nThreshold , \
+                                        const double & alphaIn , \
+                                        const double & thickness) const
+{
+    // std::cout << 0 << std::endl;
+    if (this->isEscapeCacheEnabled() == 0)
+    {
+        return false;
+    }
+    if (this->detectorEscapeCache.size() > 0)
+    {
+        // std::cout << 1 << std::endl;
+        if (energyThreshold == this->detectorEnergyThresholdUsedInCache)
+        {
+            // std::cout << 2 << std::endl;
+            if (intensityThreshold == this->detectorIntensityThresholdUsedInCache)
+            {
+                // std::cout << 3 << std::endl;
+                // std::cout << " input " << nThreshold << " cached " << this->detectorNThresholdUsedInCache << std::endl;
+                if(nThreshold == this->detectorNThresholdUsedInCache)
+                {
+                    // std::cout << 4 << std::endl;
+                    if (alphaIn == this->detectorAlphaInUsedInCache)
+                    {
+                        //std::cout << 5 << std::endl;
+                        //std::cout << " input " << thickness << " cached " << this->detectorThicknessUsedInCache << std::endl;
+                        if (thickness == this->detectorThicknessUsedInCache)
+                        {
+                            //std::cout << 6 << std::endl;
+                            // we have to compare the composition
+                            if (composition.size() == this->detectorCompositionUsedInCache.size())
+                            {
+                                // std::cout << 7 << std::endl;
+                                if (std::equal(composition.begin(), composition.end(), this->detectorCompositionUsedInCache.begin()))
+                                {
+                                    // std::cout << 8 << std::endl;
+                                    return true;
+                                }
+                            }
+                        }
+                    }
+                }
+            }
+        }
+    }
+    return false;
+}
+
+void Elements::updateEscapeCache(\
+                                        const std::map<std::string, double> & composition,
+                                        const std::vector<double> & energyList, \
+                                        const double & energyThreshold, \
+                                        const double & intensityThreshold, \
+                                        const int & nThreshold , \
+                                        const double & alphaIn , \
+                                        const double & thickness)
+{
+    std::vector<double>::size_type i;
+    double energy;
+
+    if (this->isEscapeCacheEnabled() == 0)
+    {
+        std::cout << "WARNING: Filling escape cache when escape cache is disabled" << std::endl;
+    }
+
+    if (!this->isEscapeCacheCompatible(composition, energyThreshold, \
+                                        intensityThreshold, nThreshold, alphaIn, thickness))
+    {
+        // not compatible, clear the detector cache
+        this->clearEscapeCache();
+    }
+    for (i = 0; i < energyList.size(); ++i)
+    {
+        energy = energyList[i];
+        if (this->detectorEscapeCache.find(energy) == this->detectorEscapeCache.end())
+        {
+            // std::cout << "filling energy " << energy << std::endl;
+            this->detectorEscapeCache[energy] = this->getEscape(composition, energy, energyThreshold, \
+                                                                intensityThreshold, nThreshold, alphaIn, thickness);
+        }
+        this->detectorCompositionUsedInCache = composition;
+        this->detectorEnergyThresholdUsedInCache = energyThreshold;
+        this->detectorIntensityThresholdUsedInCache = intensityThreshold;
+        this->detectorNThresholdUsedInCache = nThreshold;
+        // std::cout << "Storing cache " << nThreshold << std::endl;
+        this->detectorAlphaInUsedInCache = alphaIn;
+        this->detectorThicknessUsedInCache = thickness;
+    }
+}
+
+
+const int Elements::isCacheEnabled(const std::string & elementName) const
+{
+    std::map<std::string, int>::const_iterator it;
+    int i;
+    if (this->isElementNameDefined(elementName))
+    {
+        it = this->elementDict.find(elementName);
+        i = it->second;
+        return this->elementList[i].isCacheEnabled();
+    }
+    else
+        throw std::invalid_argument("Invalid element: " + elementName);
+}
+
+int Elements::getCacheSize(const std::string & elementName) const
+{
+    std::map<std::string, int>::const_iterator it;
+    int i;
+    if (this->isElementNameDefined(elementName))
+    {
+        it = this->elementDict.find(elementName);
+        i = it->second;
+        return this->elementList[i].getCacheSize();
+    }
+    else
+        throw std::invalid_argument("Invalid element: " + elementName);
+}
+
+
 } // namespace fisx
